"""Configuration information for iotile-ext-cloud."""

import getpass
import sys

from iotile.core.dev.registry import ComponentRegistry
from iotile.cloud.cloud import IOTileCloud
from iotile.core.utilities.typedargs import param
from iotile.core.exceptions import ArgumentError
from iotile_cloud.api.connection import Api
from builtins import input

@param("username", "string", desc="IOTile cloud username")
@param("password", "string", desc="IOTile cloud password")
def link_cloud(self, username=None, password=None, device_id=None):
    """Create and store a token for interacting with the IOTile Cloud API.

    You will need to call link_cloud once for each virtualenv that
    you create and want to use with any api calls that touch iotile cloud.

    Note that this method is called on a ConfigManager instance

    If you do not pass your username or password it will be prompted from
    you securely on stdin.

    If you are logging in for a user, the token will expire periodically and you
    will have to relogin.

    If you pass a device_id, you can obtain a limited token for that device
    that will never expire, assuming you have access to that device.

    Args:
        username (string): Your iotile.cloud username.  This is prompted
            from stdin if not provided.
        password (string): Your iotile.cloud password.  This is prompted
            from stdin if not provided.
        device_id (int): Optional device id to obtain permanent credentials
            for a device.
    """

    reg = ComponentRegistry()

    domain = self.get('cloud:server')

    if username is None:
        # Both python 2 and 3 require native strings to be passed into getpass
        prompt_str = "Please enter your IOTile.cloud email: "
        if sys.version_info.major < 3:
            prompt_str = prompt_str.encode('utf-8')

        username = input(prompt_str)

    if password is None:
        # Both python 2 and 3 require native strings to be passed into getpass
        prompt_str = "Please enter your IOTile.cloud password: "
        if sys.version_info.major < 3:
            prompt_str = prompt_str.encode('utf-8')

        password = getpass.getpass(prompt_str)

    cloud = Api(domain=domain)
    ok_resp = cloud.login(email=username, password=password)
    if not ok_resp:
        raise ArgumentError("Could not login to iotile.cloud as user %s" % username)

    reg.set_config('arch:cloud_user', cloud.username)
    reg.set_config('arch:cloud_token', cloud.token)
    reg.set_config('arch:cloud_token_type', cloud.token_type)

    if device_id is not None:
        cloud = IOTileCloud()
        cloud.impersonate_device(device_id)


def get_variables():
    """Get a dictionary of configuration variables."""

    prefix = "cloud"

    conf_vars = []
    conf_vars.append(["server", "string", "The domain name to talk to for iotile.cloud operations (including https:// prefix)", 'https://iotile.cloud'])

    return prefix, conf_vars



import pytest
from builtins import range
from iotile.core.dev.semver import SemanticVersion
from iotile.cloud.apps import OtaUpdater


def test_ota_app_creation(ota_cloud, simple_hw):

    cloud, _proj_id, _server = ota_cloud

    hw = simple_hw
    app_tag = 123
    app_version = SemanticVersion(0,0,1)
    os_tag = 234
    os_version = SemanticVersion(0,0,1)
    app_info = [app_tag, app_version]
    os_info = [os_tag, os_version]

    device_id = 1
    hw.connect(device_id)
    ota_updater = OtaUpdater(hw, app_info, os_info, device_id)
    assert ota_updater is not None


def test_no_deployments(ota_cloud, simple_hw):
    cloud, _proj_id, _server = ota_cloud
    hw = simple_hw
    hw.connect(3)

    ota_updater = OtaUpdater(hw, [123, SemanticVersion(0,0,1)], [234, SemanticVersion(0,0,1)], 3)

    actual_result = ota_updater.check_cloud_for_script()

    assert not actual_result


def test_apply_update(ota_cloud, simple_hw):

    cloud, _proj_id, _server = ota_cloud

    hw = simple_hw
    app_tag = 123
    app_version = SemanticVersion(0,0,2)
    os_tag = 234
    os_version = SemanticVersion(0,0,2)
    app_info = [app_tag, app_version]
    os_info = [os_tag, os_version]

    device_id = 1
    hw.connect(device_id)
    ota_updater = OtaUpdater(hw, app_info, os_info, device_id)

    blob = b'\x8e\x9a\x1fG\x8e)\x92\xc1\xbaz\x9e\xd3\x9a@\x86\x10L=.\x1f\x1a\x05\x00\x00\x0e\x00\x00\x00\x04\x00\x00\x00\x05 \x08\x08\x00\x00\x0c\x00\x00\x00\x04\x00\x00\x00\x0c \x08\x08\x0c\x00\x00\x00\x04\x00\x00\x00\r \x08\x08 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00D\x028\xff\xff\x00\n\n\x01\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x00\x00\x00\x00\x00\x00\x00\x00\x01D\x038\xff\xff\x00\n\n\x01\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14\x01<\xff\xff\x00\n\n\x01\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x00\x00\x00\x00\x00\x00\x00\x00\x01\x14\x02<\xff\xff\x00\n\n\x01\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x00\x00\x00\x00\x00\x00\x00\x00\x04\x14\x008\xff\xff\x00\n\n\x01\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x00\x00\x00\x00\x08\x00\x00\x00\x01$\x02$\x00\x14\x00\n\x08\x00\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x00\x00\x00\x00\x08\x00\x00\x00\x01$\x03$\x01\x14\x00\n\x08\x00\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x01\x00\x00\x00\x00\x00\x00\x00\x01P\x00D\x00$\x06\x05\n\x00\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x01\x00\x00\x00\x00\x00\x00\x00\x03\x14\x00\x14\x05$\x05\t\n\x00\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x00\x00\x00\x00h\x01\x00\x00\x00X\x04\x14\x00D\x00\n\x05\x00\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x00\x00\x00\x00\x01\x00\x00\x00\x05D\x01D\x01$\x00\n\x08\x00\x00\x00 \x00\x00\x00\x04\x00\x00\x00\x03 \x08\x08\x01\x00\x00\x00\x00\x00\x00\x00\x0b\x10\x05D\x06$\x06\x05\n\x00\x00\x00\x1a\x00\x00\x00\x04\x00\x00\x00\x07 \x08\x08\x00\x00\x00\x00\x00\x00\x00\x02\xff\xd7\x00\x01\x02\x00\x1a\x00\x00\x00\x04\x00\x00\x00\x07 \x08\x08\x00\x00\x00\x00\x00\x00\x00\x02\xff_\x80\x01\x02\x00\x1a\x00\x00\x00\x04\x00\x00\x00\x07 \x08\x08\x00\x00\x00\x00\x00\x00\x00\x02\x0b\x10\x01\x00\x02\x00\x12\x00\x00\x00\x04\x00\x00\x00\x00 \x08\x082\x80\x0b\x00\x00$\x12\x00\x00\x00\x04\x00\x00\x00\x00 \x08\x08\x00\x00\x00\x00\x01$\x12\x00\x00\x00\x04\x00\x00\x00\x00 \x08\x08\x01\x00\x00\x00\x02$\x12\x00\x00\x00\x04\x00\x00\x00\x00 \x08\x08\x00\x00\x00\x00\x03$\x12\x00\x00\x00\x04\x00\x00\x00\x00 \x08\x08\x00\x00\x00\x00\x05$\x12\x00\x00\x00\x04\x00\x00\x00\x00 \x08\x082\x80\x0b\x00\x06$\x0c\x00\x00\x00\x04\x00\x00\x00\x0e \x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08\x00 \x00\x00\x00\x00\x00\x00\x00\x02\x10\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x01\x00\x00\x00\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x082\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\n\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08B\x80\x01\x00\x00\x00\x00\x00\x00\x01\x0e\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x90\x01\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08P\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x03\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08Q\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x08\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08R\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x08\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08S\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x08\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08T\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x08\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08U\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x08\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08V\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x08\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08W\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x08\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08X\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x08\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08Y\x80\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x08\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08\x16\x00\x00\x00\x04\x00\x00\x00\x07*\x08\x08\x01\x90\x01\x00\x00\x00\x00\x00\x00\x01\r\x00\x00\x00\x04\x00\x00\x00\x08*\x08\x08\x00\x0c\x00\x00\x00\x04\x00\x00\x00\t*\x08\x08'

    try:
        ota_updater._apply_ota_update(blob=blob)
    except:
        pytest.fail("Unexpected error")


def test_inform_cloud(ota_cloud, simple_hw):

    cloud, _proj_id, _server = ota_cloud

    hw = simple_hw
    app_tag = 123
    app_version = SemanticVersion(0,0,2)
    os_tag = 234
    os_version = SemanticVersion(0,0,2)
    app_info = [app_tag, app_version]
    os_info = [os_tag, os_version]

    device_id = 1
    hw.connect(device_id)
    ota_updater = OtaUpdater(hw, app_info, os_info, device_id)

    try:
        ota_updater._inform_cloud(device_id, 'd--0000-0000-0000-0001', True)
    except:
        pytest.fail("inform cloud test failed")


def test_all_attributes(ota_cloud, simple_hw):

    cloud, _proj_id, _server = ota_cloud

    app_tags = [122, 123, 124]
    app_versions = [SemanticVersion(0, 0, 1), SemanticVersion(0, 0, 2), SemanticVersion(0, 0, 3)]

    os_tags = [233, 234, 235]
    os_versions = [SemanticVersion(0, 0, 1), SemanticVersion(0, 0, 2), SemanticVersion(0, 0, 3)]

    devices = [1, 4, 6]
    expected_results = [False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, (3, 'bar.com/zipzap'), False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, False, False,
                        (3, 'bar.com/zipzap'), False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, (1, 'bar.com/zipzap'), False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, (5, 'bar.com/zipzap'), (7, 'bar.com/zipzap'),
                        (9, 'bar.com/zipzap'), False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, (5, 'bar.com/zipzap'), (7, 'bar.com/zipzap'), (9, 'bar.com/zipzap'), False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False, False, False, False, False, False,
                        False, False, False, False, False]

    stage = 0

    for app_tag in app_tags:
        for app_version in app_versions:
            for os_tag in os_tags:
                for os_version in os_versions:
                    for device_id in devices:

                        app_info = [app_tag, app_version]
                        os_info = [os_tag, os_version]
                        hw = simple_hw
                        try:
                            hw.disconnect()
                        except:
                            pass
                        hw.connect(device_id)

                        ota_updater = OtaUpdater(hw, app_info, os_info, device_id)
                        actual_result = ota_updater.check_cloud_for_script()

                        if not actual_result:
                            print("os_tag={2}, app_tag={0}, os_version={3}, app_version={1}".format(app_tag, app_version, os_tag, os_version))
                        if actual_result:
                            print("---------")
                            print("Match achieved for device ID ", device_id)
                            print("os_tag={2}, app_tag={0}, os_version={3}, app_version={1}".format(app_tag, app_version, os_tag, os_version))
                            print("---------")

                        assert expected_results[stage] == actual_result
                        stage += 1



"""Common utility functions used across iotile.cloud packages."""

from iotile_cloud.utils.gid import IOTileDeviceSlug, IOTileFleetSlug
from iotile.core.exceptions import ArgumentError
from past.builtins import basestring

def device_slug_to_id(slug):
    """Convert a d-- device slug to an integer.

    Args:
        slug (str): A slug in the format d--XXXX-XXXX-XXXX-XXXX

    Returns:
        int: The device id as an integer

    Raises:
        ArgumentError: if there is a malformed slug
    """

    if not isinstance(slug, basestring):
        raise ArgumentError("Invalid device slug that is not a string", slug=slug)

    try:
        device_slug = IOTileDeviceSlug(slug, allow_64bits=False)
    except ValueError:
        raise ArgumentError("Unable to recognize {} as a device id".format(slug))

    return device_slug.get_id()


def device_id_to_slug(id):
    """ Converts a device id into a correct device slug.

    Args:
        id (long) : A device id
        id (string) : A device slug in the form of XXXX, XXXX-XXXX-XXXX, d--XXXX, d--XXXX-XXXX-XXXX-XXXX
    Returns:
        str: The device slug in the d--XXXX-XXXX-XXXX-XXXX format
    Raises:
        ArgumentError: if the ID is not in the [1, 16**12] range, or if not a valid string
    """

    try:
        device_slug = IOTileDeviceSlug(id, allow_64bits=False)
    except ValueError:
        raise ArgumentError("Unable to recognize {} as a device id".format(id))

    return str(device_slug)


def fleet_id_to_slug(id):
    """ Converts a fleet id into a correct fleet slug.

    Args:
        id (long) : A fleet id
        id (string) : A device slug in the form of XXXX, XXXX-XXXX-XXXX, g--XXXX, g--XXXX-XXXX-XXXX
    Returns:
        str: The device slug in the g--XXXX-XXXX-XXX format
    Raises:
        ArgumentError: if the ID is not in the [1, 16**12] range, or if not a valid string
    """

    try:
        fleet_slug = IOTileFleetSlug(id)
    except ValueError:
        raise ArgumentError("Unable to recognize {} as a fleet id".format(id))

    return str(fleet_slug)



import pytest
import requests_mock
import datetime
from dateutil.tz import tzutc
from iotile.cloud.cloud import IOTileCloud
from iotile.cloud.config import link_cloud
from iotile.core.dev.config import ConfigManager
from iotile.core.dev.registry import ComponentRegistry
from iotile.core.exceptions import ArgumentError, ExternalError
import json

@pytest.fixture
def registry():
    reg = ComponentRegistry()
    reg.clear()

    yield reg

    reg.clear()


def test_login(registry):
    """Make sure successful login is properly handled."""

    payload = {
        'jwt': 'big-token',
        'username': 'user1'
    }

    manager = ConfigManager()

    with requests_mock.Mocker() as mocker:
        mocker.post('https://iotile.cloud/api/v1/auth/login/', json=payload)

        link_cloud(manager, 'user1@random.com', 'password')

    assert registry.get_config('arch:cloud_user') == 'user1'
    assert registry.get_config('arch:cloud_token') == 'big-token'


def test_refresh(registry):
    """Make sure we can properly refresh our jwt token."""

    payload = {
        'token': 'new-token'
    }

    registry.set_config("arch:cloud_token", 'old-token')
    registry.set_config("arch:cloud_user", 'test_user')

    cloud = IOTileCloud()

    with requests_mock.Mocker() as mocker:
        mocker.post('https://iotile.cloud/api/v1/auth/api-jwt-refresh/', json=payload)

        cloud.refresh_token()

    assert registry.get_config('arch:cloud_token') == 'new-token'


def test_alternative_domains(registry):
    """Make sure we can specify an alternative domain."""

    payload = {
        'jwt': 'big-token',
        'username': 'user1'
    }

    manager = ConfigManager()
    manager.set('cloud:server', 'https://testcloud.com')

    with requests_mock.Mocker() as mocker:
        mocker.post('https://testcloud.com/api/v1/auth/login/', json=payload)

        link_cloud(manager, 'user1@random.com', 'password')

    assert registry.get_config('arch:cloud_user') == 'user1'
    assert registry.get_config('arch:cloud_token') == 'big-token'

    cloud = IOTileCloud()

    payload = {
        'token': 'new-token'
    }

    with requests_mock.Mocker() as mocker:
        mocker.post('https://testcloud.com/api/v1/auth/api-jwt-refresh/', json=payload)

        cloud.refresh_token()

    assert registry.get_config('arch:cloud_token') == 'new-token'


def test_check_time():
    """ Make sure we can check if the time is correct"""

    json_true = {'now': datetime.datetime.now(tzutc()).strftime('%a, %d %b %Y %X %Z')}
    json_false = {'now': 'Wed, 01 Sep 2010 17:30:32 GMT'}
    payload = {
        'jwt': 'big-token',
        'username': 'user1'
    }

    manager = ConfigManager()

    with requests_mock.Mocker() as mocker:

        mocker.post('https://iotile.cloud/api/v1/auth/login/', json=payload)
        link_cloud(manager, 'user1@random.com', 'password')
        cloud = IOTileCloud()
        mocker.get('https://iotile.cloud/api/v1/server/', json=json_true)
        assert cloud.check_time() == True
        mocker.get('https://iotile.cloud/api/v1/server/', json=json_false)
        assert cloud.check_time() == False


def test_get_fleet():
    """Make sure we can get fleets."""

    auth_payload = {
        'jwt': 'big-token',
        'username': 'user1'
    }
    test_payload = {"count":1,
              "next":"Null",
              "previous":"Null",
              "results":[{"device":"d--0000-0000-0000-0001","always_on":True,"is_access_point":False}]}

    expected = {
        "d--0000-0000-0000-0001":{
            "always_on":True,
            "is_access_point":False}
    }
    manager = ConfigManager()
    with requests_mock.Mocker() as mocker:

        mocker.post('https://iotile.cloud/api/v1/auth/login/', json=auth_payload)
        link_cloud(manager, 'user1@random.com', 'password')
        cloud = IOTileCloud()
        mocker.get('https://iotile.cloud/api/v1/fleet/g--0000-0000-0001/devices/', json=test_payload)
        mocker.get('https://iotile.cloud/api/v1/fleet/g--0000-0000-0002/devices/', status_code=404)
        assert cloud.get_fleet(1) == expected
        with pytest.raises(ArgumentError):
            cloud.get_fleet(2)
        with pytest.raises(ArgumentError):
            cloud.get_fleet(pow(16,12) + 1)


def test_get_whitelist():
    """ Make sure we can retrieve the whitelist correctly """
    with open('test/large_mock_answer.json') as lma:
        j = json.load(lma)
        test_payload = j['whitelist_test']
        p1 = j['whitelist_g1']
        p2 = j['whitelist_g2']
        p3 = j['whitelist_g3']
        expected = j['expected']
        empty_whitelist_test = j['empty_whitelist_test']
        p4 = j['whitelist_g4']
    payload = {
        'jwt': 'big-token',
        'username': 'user1'
    }

    manager = ConfigManager()
    with requests_mock.Mocker() as mocker:

        mocker.post('https://iotile.cloud/api/v1/auth/login/', json=payload)
        link_cloud(manager, 'user1@random.com', 'password')
        cloud = IOTileCloud()
        mocker.get('https://iotile.cloud/api/v1/fleet/?device=d--0000-0000-0000-0001', status_code=404)
        with pytest.raises(ExternalError):
            cloud.get_whitelist(1)
        mocker.get('https://iotile.cloud/api/v1/fleet/?device=d--0000-0000-0000-0002', json={'results':[]})
        with pytest.raises(ExternalError):
            cloud.get_whitelist(2)
        mocker.get('https://iotile.cloud/api/v1/fleet/?device=d--0000-0000-0000-01bd', json=test_payload)
        mocker.get('https://iotile.cloud/api/v1/fleet/g--0000-0000-0001/devices/', json=p1)
        mocker.get('https://iotile.cloud/api/v1/fleet/g--0000-0000-0002/devices/', json=p2)
        mocker.get('https://iotile.cloud/api/v1/fleet/g--0000-0000-0003/devices/', json=p3)
        assert cloud.get_whitelist(0x1bd) == expected
        mocker.get('https://iotile.cloud/api/v1/fleet/?device=d--0000-0000-0000-01bd', json=empty_whitelist_test)
        mocker.get('https://iotile.cloud/api/v1/fleet/g--0000-0000-0004/devices/', json=p4)
        with pytest.raises(ExternalError):
            cloud.get_whitelist(0x1bd)




from setuptools import setup, find_packages

import version

setup(
    name="iotile-ext-cloud",
    packages=find_packages(exclude=("test",)),
    version=version.version,
    license="LGPLv3",
    install_requires=[
        "iotile-core>=4.0.0",
        "iotile_cloud>=0.9.9"
    ],
    python_requires=">=3.5,<4",
    entry_points={'iotile.config_function': ['link_cloud = iotile.cloud.config:link_cloud'],
                  'iotile.config_variables': ['iotile-ext-cloud = iotile.cloud.config:get_variables'],
                  'iotile.plugin': ['cloud = iotile.cloud.plugin:setup_plugin'],
                  'iotile.app': ['cloud_uploader = iotile.cloud.apps.cloud_uploader',
                                 'ota_updater = iotile.cloud.apps.ota_updater']},
    description="IOTile.cloud integration into CoreTools",
    author="Arch",
    author_email="info@arch-iot.com",
    url="http://github.com/iotile/lib_iotilecore",
    keywords=["iotile", "arch", "embedded", "hardware", "firmware"],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries :: Python Modules"
        ],
    long_description="""\
IOTile Cloud Extensions
-----------------------

A set of extensions to IOTile CoreTools that allow you to interact with iotile.cloud.
See https://www.arch-iot.com.
"""
)



"""An IOTileApp that uploads reports from an iotile device to the cloud."""

import logging
import time
import struct
from builtins import range
from future.utils import viewitems
from iotile.core.exceptions import HardwareError, ArgumentError
from iotile.core.hw import IOTileApp
from iotile.core.hw.reports import SignedListReport
from iotile.core.utilities.console import ProgressBar
from iotile.cloud import IOTileCloud, device_id_to_slug
from typedargs.annotate import docannotate, context


@context("CloudUploader")
class CloudUploader(IOTileApp):
    """An IOtile app that can get reports from a device and upload them to the cloud.

    This app performs the same basic upload functionality as the IOTile Companion
    mobile application.  It has a single important method: upload.  This method
    will start a loop that:
        - acknowledges old data from the device that has safely reched iotile.cloud
        - triggers the device to send all of its data
        - waits for all data to be sent
        - uploads all reports to iotile.cloud

    Args:
        hw (HardwareManager): A HardwareManager instance connected to a
            matching device.
        app_info (tuple): The app_tag and version of the device we are
            connected to.
        os_info (tuple): The os_tag and version of the device we are
            connected to.
        device_id (int): The UUID of the device that we are connected to.
    """

    logger = logging.getLogger(__name__)

    def __init__(self, hw, app_info, os_info, device_id):
        super(CloudUploader, self).__init__(hw, app_info, os_info, device_id)

        self._cloud = IOTileCloud()
        self._con = self._hw.get(8, basic=True)

    @classmethod
    def AppName(cls):
        return 'cloud_uploader'

    def _get_uuid(self):
        device_id, = self._con.rpc(0x10, 0x08, result_format="L16x")
        return device_id

    def _trigger_streamer(self, index):
        error, = self._con.rpc(0x20, 0x10, index, result_format="L")

        # This error code means that the streamer did not have any new data
        if error == 0x8003801f:
            self.logger.info("We manually triggered streamer %d but it reported that there were no new readings", index)
        elif error != 0:
            raise HardwareError("Error triggering streamer", code=error, index=index)

    def _streamer_finished(self, index):
        res = self._con.rpc(0x20, 0x0a, index, result_type=(0, True))

        #Check if we got an error that means the streamer doesn't exist
        if len(res['buffer']) == 4:
            _error, = struct.unpack("<L", res['buffer'])
            return None

        comm_status, = struct.unpack("<18xBx", res['buffer'])
        return comm_status == 0

    def _wait_streamers_finished(self, timeout=60*10.0):
        start = time.time()

        while (time.time() - start) < timeout:
            for i in range(0, 16):
                self.logger.info("Waiting for streamer %d", i)
                while True:
                    status = self._streamer_finished(i)
                    if status is None:
                        self.logger.info("No streamer %d, all streamers finished", i)
                        return
                    elif status is True:
                        break

                    time.sleep(1.0)

                self.logger.info("Streamer %d finished", i)

        raise HardwareError("Device took too long to stream data", timeout_seconds=timeout)

    def _ack_streamer(self, index, value, force=False):
        if index <= 0xFF:
            error, = self._con.rpc(0x20, 0x0f, index, int(force), value, arg_format="HHL", result_format="L")

            if error == 0x8003801e:
                # This means the streamer value is older than what is already there, this is not an error
                pass
            elif error:
                raise HardwareError("Error acknowledging streamer", error_code=error, index=index, value=value)
        else:
            raise ArgumentError("Streamer index is out of bounds", index=index)

    def _wait_finished_streaming(self):
        time.sleep(1.0)
        self._wait_streamers_finished()

    def download(self, trigger=None, acknowledge=True, force=None):
        """Synchronously download all reports from the device.

        This function will:
        - acknowledge old data from the device that has safely reached iotile.cloud
          unless you pass acknowledge=False
        - trigger the device to send all of its data.  This happens automatically
          when we enable_streaming.  However, if you need to manually trigger a
          streamer, you can specify that using trigger=X where X is in the index
          of the streamer to trigger.
        - wait for all data to be received from the device.

        The only differnce between this function and upload is that this function
        will return the reports as a list, rather than uploading them directly
        to iotile.cloud.

        Args:
            trigger (int): If you need to manually trigger a streamer on the device,
                you can specify its index here and it will have trigger_streamer called
                on it before we enter the upload loop.
            acknowledge (bool): If you don't want to send all cloud acknowledgements
                down to the device before enabling streaming, you can pass False.  The
                default behavior is True.
            force (dict): If you want to forcibly set the acknowledgement for a given
                streamer you can pass it in this dictionary.  The value passed will
                be used instead of whatever was received from the cloud.  It can be
                a number, which is interpeted as the value to acknowledge, or it can
                be a tuple with a number and boolean, which is interpreted as the
                force parameter.

        Returns:
            list of IOTileReport: The list of reports received from the device.
        """

        device_id = self._get_uuid()
        slug = device_id_to_slug(device_id)

        self.logger.info("Connected to device 0x%X", device_id)

        streamer_acks = []
        if force is not None:
            for index, value in viewitems(force):
                force_ack = False
                if isinstance(value, tuple):
                    value, force_ack = value

                streamer_acks.append((index, value, force_ack))

        if acknowledge:
            self.logger.info("Getting acknowledgements from cloud for slug %s", slug)

            resp = self._cloud.api.streamer.get(device=slug)
            acks = resp.get('results', [])
            self.logger.info("Found %d acknowledgements", len(acks))

            for ack in acks:
                index = ack['index']
                last_id = ack['last_id']

                if index <= 0xFF:
                    streamer_acks.append((index, last_id, False))

        else:
            self.logger.info("Not acknowledging readings from cloud per user request")

        for index, last_id, force in streamer_acks:
            self.logger.info("Acknowledging highest ID %d for streamer %d (force=%s)", last_id, index, force)
            self._ack_streamer(index, last_id, force=force)

        # Configure Downloader to not break up the report
        self.set_report_size()  #Set to max report size
        self._hw.enable_streaming()

        if trigger is not None:
            self.logger.info("Explicitly triggering streamer %d", trigger)
            self._trigger_streamer(trigger)

        self._wait_streamers_finished()

        reports = [x for x in self._hw.iter_reports()]
        signed_reports = [x for x in reports if isinstance(x, SignedListReport)]

        self.logger.info("Received %d signed reports, ignored %d realtime reports", len(signed_reports), len(reports) - len(signed_reports))

        return signed_reports

    @docannotate
    def upload(self, trigger=None, acknowledge=True):
        """Synchronously get all data from the device and upload it to iotile.cloud.

        This function will:
        - acknowledge old data from the device that has safely reached iotile.cloud
          unless you pass acknowledge=False
        - trigger the device to send all of its data.  This happens automatically
          when we enable_streaming.  However, if you need to manually trigger a
          streamer, you can specify that using trigger=X where X is in the index
          of the streamer to trigger.
        - wait for all data to be received from the device.
        - upload all reports to iotile.cloud securely.

        If you want to see details about what is happening, you can capture the
        logging output.

        This method will use whatever the default iotile.cloud domain and credentials
        are that are configured in your current virtualenv.

        Args:
            trigger (int): If you need to manually trigger a streamer on the device,
                you can specify its index here and it will have trigger_streamer called
                on it before we enter the upload loop.
            acknowledge (bool): If you don't want to send all cloud acknowledgements
                down to the device before enabling streaming, you can pass False.  The
                default behavior is True.
        """

        signed_reports = self.download(trigger, acknowledge)

        for report in signed_reports:
            self.logger.info("Uploading report with ids in (%d, %d)", report.lowest_id, report.highest_id)
            self._cloud.upload_report(report)

    @docannotate
    def get_report_size(self):
        """ Sets and verifies the report size for a pod
        Returns:
           int: The maximum size of a report
        """
        maxpacket, _comp1, comp2, = self._con.rpc(0x0A, 0x06, result_format="LBB")
        return maxpacket

    @docannotate
    def set_report_size(self, size=0xFFFFFFFF):
        """ Sets and verifies the report size for a pod
        Args:
            size (int): The maximum size of a report
        """

        error, = self._con.rpc(0x0A, 0x05, size, 0, arg_format="LB", result_format="L")

        if error:
            raise HardwareError("Error setting report size.", error_code=error, size=size)

        maxpacket, _comp1, comp2, = self._con.rpc(0x0A, 0x06, result_format="LBB")

        if maxpacket != size:
            raise HardwareError("Max Packet Size was not set as expected")



